//blaster.cpp

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <qrect.h>
#include <qpoint.h>
#include <qimage.h>
//Added by qt3to4:
#include <QPixmap>

#include <QList>

#include "mainwin.h"
#include "blaster.h"
#include "bunny.h"
#include "errors.h"
#include "dbbconfig.h"
#include "logger.h"

#define UNUSED( x ) x = x

//#define DEBUGOUT

#ifdef DEBUGOUT

#include <iostream>
#define cout std::cout
#define endl std::endl

#endif //DEBUGOUT

//This is the difference in size between the inner rect 
//enclosing the bunny and the outer rect. We use this value
//when constructing the outer rect and when calculating the
//x and y shifts for bunnies that lie near or outside the 
//image border.  
const int rectdelta = 2;

extern mainwin * pMainWin;

blaster::blaster()
{
#ifdef DEBUGOUT
	cout << "constructing blaster" << endl;
#endif
	for (int i = 0; i < ((int)(sizeof(rgbhilos) / sizeof(int))); i++)
	{
		rgbhilos[i] = -1;
	}
	rgbSum = 0;
	
	pLogEdit = 0;
	fpLogFile = 0;
}
  
blaster::~blaster()
{
#ifdef DEBUGOUT
	cout << "destroying blaster" << endl;
#endif
}

// Calculates the rect containing the bunny. 
// Returns 0 if there is an error, 1 otherwise.
// If the bunny is near the edge of the QPixmap, the rect will NOT be shifted 
// so that it is entirely contained within the QPixmap, but we do store the amount
// that would be needed so that the actual painting routines can ignore invalid pixels.
int blaster::getBunnyRect(bunnies * pbl,  // address of bunnies list
                          int nbunny,     // 0 based index into bunnies list
                          QRect & rc,     // Will contain the rect coordinates containing the bunny
                          int maxright,   // width of the jpeg image QPixmap
                          int maxbottom)  // height of the jpeg image QPixmap
{
	if (pbl == 0) {
		ERROR(ERR_NULL_OBJECT, "blaster::getBunnyRect(), pbl is NULL");
		return ERR_NULL_OBJECT;
	}
	
	bunny * pb = pbl->getAt(nbunny);
	if (pb == 0) {
		ERROR(ERR_NULL_OBJECT, "blaster::getBunnyRect(), pb is NULL");
		return ERR_NULL_OBJECT;
	}
	
	int radius = ( (pb->getDia() / 2) ); 
	QPoint ptTopLeft( pb->getX() - radius, pb->getY() - radius );
	QPoint ptBottomRight( pb->getX() + radius, pb->getY() + radius );
	
	rc.setTopLeft(ptTopLeft);
	rc.setBottomRight(ptBottomRight);
	
	//Calculate and save the xshift and yshift
	int xshift = 0;
	int yshift = 0;
	getXYShift( rc, maxright, maxbottom, &xshift, &yshift );
	
	pb->setXShift(xshift);
	pb->setYShift(yshift);
	
#ifdef DEBUGOUT
	cout << "image maxbottom=" << maxbottom << " maxright=" << maxright << endl;
	cout << "bunny x=" << pb->getX() \
	<< " y=" << pb->getY() \
	<< " bunny radius=" << pb->getDia() / 2 << endl;
	
	cout << "bunny rect: left=" << rc.left() \
	<< " top="  << rc.top() \
	<< " right=" << rc.right() \
	<< " bottom=" << rc.bottom() \
	<< " xshift=" << pb->getXShift() \
	<< " yshift=" << pb->getYShift() << endl;
#endif  
	
	return ERR_NOERR;
}


// Calculate a rectangle that surrounds the inner bunny rect.
// We will use pixel RGB values in the outer rect to interpolate the pixels
// used to erase the dust bunny. This rect gets clipped if the bunny lies 
// near the edge of the image. See getXYShift().
int blaster::getSurroundingRect(QRect & inner, // rect containing the bunny
                                QRect & rc)    // will contain the outer rect coords
{
	if (inner.isNull() || !inner.isValid()) {
		ERROR(ERR_NULL_OBJECT, "getSurroundingRect(), inner rect is NULL or invalid");
		return ERR_NULL_OBJECT;
	}
	
	int left = inner.left() - rectdelta;
	int top = inner.top() -rectdelta;
	int width = inner.width() + (2*rectdelta);
	
	rc.setLeft(left);
	rc.setTop(top);
	rc.setWidth(width);
	rc.setHeight(width);
	
#ifdef DEBUGOUT	
	cout << "outer rect: left=" << rc.left() \
	<< " top="  << rc.top() \
	<< " right=" << rc.right() \
	<< " bottom=" << rc.bottom() \
	<< endl;
#endif

	return ERR_NOERR;
}


//Calculates the average brightness and average gradient of the 
//pixel rgb values of the pixels used to interpolate the rgb value
//used to erase the bunny.
//The idea is to determine when we want to avoid erasing a bunny,
//such as when it is in a dark area of the image, or in an area 
//that has a steep color gradient. In a dark area you dont see the
//bunny anyway, and in a high gradient area the interpolate routine
//tends to blur image detail.
int blaster::calcRgbGradients(QImage & img,	//Contains the bunny rect + outer rect  
							  int innerradius, 	//bunny rect radius							  
							  int outerradius, 	//outer rect radius
							  int xshift, 		//xshift if the bunny lies near the image border
							  int yshift)		//yshift if the bunny lies near the image border
{
	int row, col, radius;
	QPoint ptcenter = img.rect().center();
	int centerx = ptcenter.x();
	int centery = ptcenter.y();
	QRgb rgb;
	int maxred, minred, maxgreen, mingreen, maxblue, minblue;
	int red, green, blue;
	
	maxred = maxgreen = maxblue = 0;
	minred = mingreen = minblue = 512;
	
	rgbSum = (long)0;
	avgrgbrange = 0;
	
	long redsum = (long)0;
	long greensum = (long)0;
	long bluesum = (long)0;
	int count = 0;
	
	//adjust for x/y shift
	//otherwise we include pixels outside the image
	//which throws off the gradient calculations
	int colstart, rowstart, colend, rowend;
	colstart = rowstart = 0;
	colend = img.width();
	rowend = img.height();
	
	if (xshift > 0) colstart = xshift;
	if (yshift > 0) rowstart = yshift;
	if (xshift < 0) colend = img.width() + xshift;
	if (yshift < 0) rowend = img.height() + yshift;
	
	for (row = rowstart; row < rowend; row++)
	{
		for (col = colstart; col < colend; col++)
		{
			radius = blaster::radius(centerx, centery, col, row); 
			if (radius > innerradius && radius < outerradius)
			{
				rgb = img.pixel(col, row);
				
				red = qRed(rgb);
				green = qGreen(rgb);
				blue = qBlue(rgb);
				
				count++;
				redsum += red;
				greensum += green;
				bluesum += blue;
				rgbSum += (red+green+blue);
				
				maxred = qMax(maxred, red);
				minred = qMin(minred, red);
				
				maxgreen = qMax(maxgreen, green);
				mingreen = qMin(mingreen, green);
				
				maxblue = qMax(maxblue, blue);
				minblue = qMin(minblue, blue);
			}
		}
	}
	
	rgbhilos[0] = maxred;
	rgbhilos[1] = minred;
	rgbhilos[2] = maxgreen;
	rgbhilos[3] = mingreen;
	rgbhilos[4] = maxblue;
	rgbhilos[5] = minblue;
	
	avgrgbrange = ( (maxred-minred) + (maxgreen-mingreen) + (maxblue-minblue) ) / 3;
	rgbSum = (rgbSum/count)/3;
	
#ifdef DEBUGOUT  
	cout << "***calcRgbGradients()***" << endl;
	cout << "pixels included in calculation: " << count << endl;
	cout << "max/min red: " << maxred << "-" << minred << " = " << maxred - minred << endl;
	cout << "max/min green: " << maxgreen << "-" << mingreen << " = " << maxgreen - mingreen << endl;
	cout << "max/min blue: " << maxblue << "-" << minblue << " = " << maxblue - minblue << endl;
	cout << "avg gradient: " << avgrgbrange << " <---avgrgbrange" << endl;		
	
	cout << "redsum: " << redsum << " avg: " << redsum/count << endl;
	cout << "greensum: " << greensum << " avg: " << greensum/count <<endl;
	cout << "bluesum: " << bluesum << " avg: " << bluesum/count << endl;
	cout << "rgbsum: " << rgbSum << " avg: " << rgbSum << " <---brightness" << endl;
	cout << "***calcRgbGradients()***" << endl;
#endif  
	
  return ERR_NOERR;
}


//Erases a dust bunny by interpolation from the pixels just ouside the bunny
//Since it requires a QPixmap object, it cant be called from a non GUI program.
int blaster::interpolate(bunnies * pbl,		//pointer to the bunny list 
						 int nbunny, 		//index into the list, the bunny we are going to erase
						 QPixmap * pmjpg)	//the entire image pixmap
{
	QRect rcinner, rcouter;
	
	if (getBunnyRect(pbl, nbunny, rcinner, pmjpg->width(), pmjpg->height()) != ERR_NOERR ) return 0;
	if (getSurroundingRect(rcinner, rcouter) != ERR_NOERR) return 0;
	
	//Create a pixmap the size of the outer bunny rect
	QPixmap *pdest = new QPixmap(rcouter.width(), rcouter.height());
	
	//Copy the image rect containing the bunny to the new pixmap
	copyBlt(pdest, 0, 0, pmjpg, rcouter.x(), rcouter.y(), rcouter.width(), rcouter.height());
	
	//Convert to a QImage so we can get at the pixels
	QImage img = pdest->convertToImage();
	
	//blast the bunnies
	interpolateQImage( img, pbl, nbunny );
	
	//convert back to QPixmap, blit the pixmap rect back onto the large
	//jpeg image pixmap.
	pdest->convertFromImage(img, 0);
	copyBlt(pmjpg, rcouter.x(), rcouter.y(), pdest, 0, 0, rcouter.width(), rcouter.height());
	delete pdest;
	
	//The will end up removing the user drawn circle around the bunny
	pbl->remove(pbl->getAt(nbunny));
	
	return ERR_NOERR;
}

//Get the pixmap of the rectangle containing the bunny and the pixmap's
//upper left x,y coordinates relative to the image pixmap.
//We will stuff the pixmap and x,y coords into the undo buffer
//so we can undo the blasting operation.
QPixmap& blaster::getBunnyPixmap(bunnies * pbl, 	//ptr to the bunny list
								int nbunny, 		//0 based index into the bunny list
								QPixmap * pmjpg, 	//pixmap of the entire image
								QPoint& pt			//receives x,y coords of upper left corner of the
								)					//pixmap area containing the bunny
{
	QRect rcinner, rcouter;
	
	getBunnyRect(pbl, nbunny, rcinner, pmjpg->width(), pmjpg->height());
	getSurroundingRect(rcinner, rcouter);
	
	//Create a pixmap the size of the outer bunny rect
	QPixmap* pmOut = new QPixmap(rcouter.width(), rcouter.height());
	
	//Copy the image rect containing the bunny to the new pixmap
	copyBlt(pmOut, 0, 0, pmjpg, rcouter.x(), rcouter.y(), rcouter.width(), rcouter.height());
	
	//get the upper left corner coordinates. We need these so we
	//know where to restore the pixmap to later.
	pt.setX(rcouter.x());
	pt.setY(rcouter.y());
	
	return *pmOut;
}

int blaster::undoBlast(QPixmap * pmjpg, QPixmap& pmbunny, QPoint& pt)
{
	copyBlt(pmjpg, pt.x(), pt.y(), &pmbunny, 0, 0, pmbunny.width(), pmbunny.height());
	
	return ERR_NOERR;	
}


// This is the pixmap independent interpolate routine.
// NOTE: the QImage passed to this function only contains 
// the section of the image containing the bunny. 
int blaster::interpolateQImage( QImage & img,		// The QImage section containing the bunny 
								bunnies * pbl,		// Bunny list pointer
								int nbunny			// 0 based index into the bunny list
							  )
{
	int row, col, radius;
	QRgb rgb;
	QPoint ptcenter;
	ptcenter = img.rect().center();
		
	//adjust in case the bunny is at the edge of the pixmap
	int xshift = pbl->getAt(nbunny)->getXShift();
	int yshift = pbl->getAt(nbunny)->getYShift();
	
	calcRgbGradients(img, pbl->getAt(nbunny)->getDia() / 2, img.width() / 2, xshift, yshift);
	if ( !blasterThreshholdMet() )
	{
#ifdef DEBUGOUT
		cout << "threshhold not met" << endl;
#endif		
		return ERR_THRESHHOLD_NOT_MET;	
	}
	
	int colstart, rowstart, colend, rowend;
	colstart = rowstart = 0;
	colend = img.width();
	rowend = img.height();
	
	if (xshift > 0) colstart = xshift;
	if (yshift > 0) rowstart = yshift;
	if (xshift < 0) colend = img.width() + xshift;
	if (yshift < 0) rowend = img.height() + yshift;
	
	//Start by interpolating pixels that lie on the bunny radius, then progressively
	//decrease the radius.
	radius = pbl->getAt(nbunny)->getDia() / 2;
	while (radius >=  0)
	{
		for (row = rowstart; row < rowend; row++)
		{ 
			for (col = colstart; col < colend; col++)
			{
				if ( radius == blaster::radius(ptcenter.x(), ptcenter.y(), col, row) )
				{
					if (interpolatePixel(img, rgb, ptcenter.x(), ptcenter.y(), col, row, radius ) == ERR_NOERR)
					{
						if ( img.valid(col, row) )
						{
							img.setPixel(col,row, rgb);
						}
					}
				}
			}
		}
		radius--;
		colstart++;
		colend--;
		rowstart++;
		rowend--;
	}
	
	return ERR_NOERR;
}


// NOTE: the QImage passed to this function contains the entire image file data.
// This function is simmilar to interpolate(), but it can be called from a
// non GUI program, since it doesn't use a QPixmap.
int blaster::blastBunnyFromQImage(QImage & img,		//The QImage containing the entire image
								  bunnies* pbl, 	//pointer to the bunny list
								  int nbunny )		//bunny index
{
	//get the inner and outer rect
	QRect rcinner, rcouter;
	
	if (getBunnyRect(pbl, nbunny, rcinner, img.width(), img.height()) != ERR_NOERR ) return 0;
	if (getSurroundingRect(rcinner, rcouter) != ERR_NOERR) return 0;
	
	//copy the QImage section corresponding to the outer rect into a new QImage
	QImage subimg = img.copy( rcouter.left(), rcouter.top(), rcouter.width(), rcouter.height() );
	
	// interpolate, erase the bunny
	interpolateQImage( subimg, pbl, nbunny );
	
	// copy the pixels from the new QImage back to the correct location
	// in the original QImage containing the image file data.
	bitBlt( &img, rcouter.left(), rcouter.top(), &subimg, 0, 0, rcouter.width(), rcouter.height());
	
	return 1;
}


//Interpolate pixel values based on a weighted average of the 8 surrounding
//adjacent pixels which are ouside the currently used radius.
int blaster::interpolatePixel(QImage & img,		//rect containing the bunny rect plus outer rect
							  QRgb & rgb,		//will contain the interpolated rgb values
							  int centerx,		//bunny center x
							  int centery,		//bunny center y
							  int col,			//column in the outer rect
							  int row,			//row in the outer rect
							  int radius )		//bunny radius
{
	float rsum, gsum, bsum;
	rsum = gsum = bsum = 0.0;
	float nearweight, farweight, totalweight;
	QRgb interpolatedrgb;
	
	QPoint pts[8];
	pts[0].setX(col);		pts[0].setY(row-1);
	pts[1].setX(col+1);		pts[1].setY(row-1);
	pts[2].setX(col+1);		pts[2].setY(row);
	pts[3].setX(col+1);		pts[3].setY(row+1);
	pts[4].setX(col);		pts[4].setY(row+1);
	pts[5].setX(col-1);		pts[5].setY(row+1);
	pts[6].setX(col-1);		pts[6].setY(row);
	pts[7].setX(col-1);		pts[7].setY(row-1);
	
	totalweight = 0.0;
	nearweight = 1.0;
	farweight = 1.0 / 1.4142;
	
	for (int i = 0; i < 8; i++)
	{
		if  ( img.valid( pts[i].x(),pts[i].y() ) && 
			blaster::radius(centerx, centery, pts[i].x(), pts[i].y()) > radius )
		{
			interpolatedrgb = img.pixel(pts[i].x(), pts[i].y());
			
			if ( (i % 2) == 0)
			{
				rsum += ((float)(qRed(interpolatedrgb))) * nearweight;
				gsum += ((float)(qGreen(interpolatedrgb))) * nearweight;
				bsum += ((float)(qBlue(interpolatedrgb))) * nearweight;
				totalweight = totalweight + nearweight;
			}
			else
			{
				rsum += ((float)(qRed(interpolatedrgb))) * farweight;
				gsum += ((float)(qGreen(interpolatedrgb))) * farweight;
				bsum += ((float)(qBlue(interpolatedrgb))) * farweight;
				totalweight = totalweight + farweight;
			}
		}
	}
	
	if (totalweight > 0)
	{
		//round up, otherwise blasted areas will appear very slightly,
		//but noticeably darker than the surrounding area.
		rsum = (rsum / totalweight) + 0.5;
		gsum = (gsum / totalweight) + 0.5;
		bsum = (bsum / totalweight) + 0.5;
	}
	else 
	{
		ERROR(ERR_NOINTERPOLATE, "interpolatePixel(), totalweight <= 0");
		return ERR_NOINTERPOLATE;
	}
	
	rgb = qRgb((int)rsum, (int)gsum, (int)bsum);
	  
	return ERR_NOERR;
}

bool blaster::blasterThreshholdMet()
{
	
	bool bAvgColorLightnessThreshhold = false;
	bool bAvgRgbRangeThreshhold = false;
	
	if ( rgbSum > pMainWin->appSettings.avgColorLightness ) bAvgColorLightnessThreshhold = true;
	if ( avgrgbrange <= pMainWin->appSettings.avgColorRange ) bAvgRgbRangeThreshhold = true;
	
	if ( ( bAvgColorLightnessThreshhold == true && bAvgRgbRangeThreshhold == true ) ||
	       pMainWin->appSettings.bThreshholdsDisabled == true )
	{
		logger("    BLASTED");
		return true;
	}
	else
	{
		QString s2, s3;
		s2.sprintf("%ld", rgbSum);
		s3.sprintf("%d" ,avgrgbrange );
		QString s1("    NOTBLASTED AvgLighness = " +
				s2 + " AvgRange = " + s3);
		
		logger((char*)s1.latin1());
		return false;
	}
};
	
void blaster::logger(char* pstr)
{
	if (pLogEdit != 0)
	{
		LOGGER(pstr, pLogEdit);
	}
	else if (fpLogFile != 0)
	{
		LOGGER(pstr, fpLogFile);
	}
};

////////////////////////////////////////////////////////////////////////////////////////////////
//Static functions

//Static function
int blaster::radius(int inx, int iny, int outx, int outy)
{
	int x = abs(inx - outx);
	int y = abs(iny - outy);
	float root = sqrtf((float)((float)(x*x) + (float)(y*y)));
	
	//round up
	return (int)(root + 0.5);
}

//Static function 
// Bunnies that lie on the edge of the image will be contained
// in a rect that may lie partly outside the image. We need to
// calculate how far outside the image the rect extends, so that
// we can avoid interpolating pixels from outside the image.
int blaster::getXYShift(QRect & rcinner, 	//The inner bunny rect
						int maxright, 		//width of the entire image
						int maxbottom, 		//height of the entire image
						int* pxshift, 		//will receive the calculated xshift
						int* pyshift		//will receive the calculated yshift
					   )
{
	if ( (rcinner.left()-rectdelta) <= 0 ) *pxshift = abs(rcinner.left() - rectdelta)+1;
	if ( (rcinner.top()-rectdelta) <= 0 )  *pyshift = abs(rcinner.top() -rectdelta)+1;
	if ( (rcinner.right()+rectdelta) - maxright >= -1 ) *pxshift = (((rcinner.right() + rectdelta) - maxright) * -1) - 2;
	if ( (rcinner.bottom()+rectdelta) - maxbottom >= -1 ) *pyshift = (((rcinner.bottom() + rectdelta) - maxbottom) * -1) - 2;
	
	return ERR_NOERR;
}
