//
// C++ Implementation: hotpixel
//
// Description: 
//
//
// Author: ,,, <heinzg@denizen>, (C) 2007
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <stdlib.h>

#include <qimage.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qstringlist.h>

#include "errors.h"
#include "hotpixel.h"

#define UNUSED( x ) x = x

//#define DEBUGOUT

#ifdef DEBUGOUT

#include <iostream>
#define cout std::cout
#define endl std::endl

#endif //DEBUGOUT



hotpixel::hotpixel(QStringList & sList)
{
	pixelList = sList;
	initDoneList();
}

hotpixel::~hotpixel()
{
	
}

int hotpixel::processHotPixels(QImage & image)
{
	QRgb rgb;
	
	initDoneList();
	
	for (int i = 0; i < pixelList.size(); i++)
	{
		if (!pixelList.at(i).isEmpty())
		{
			//for each item in pPixelList, convert to integer coordinates
			QString s = pixelList.at(i);
			int x = getX(s);
			int y = getY(s);
			
			//interpolate the pixel
			int ret = interpolate(image, rgb, x, y);
			
			if (ret == ERR_NOERR)
			{
				QString s = "true";
				doneList.replace(i, s);;
			}
			
			//write the interpolated pixel to QImage
			if (ret == ERR_NOERR && image.valid(x,y))
			{
				image.setPixel(x, y, rgb);	
			}
		}
	}
	
	return ERR_NOERR;	
}


int hotpixel::interpolate(QImage & image,	//QImage object reperesenting the image data
						  QRgb & rgb, 		//will contain the interpolated rgb values
						  int col,			//x coordinate of the hot pixel  
						  int row)			//y coordinate of the hot pixel
{
	float rsum, gsum, bsum;
	rsum = gsum = bsum = 0.0;
	float nearweight, farweight, totalweight;
	QRgb interpolatedrgb;
	
	QPoint pts[8];
	pts[0].setX(col);		pts[0].setY(row-1);
	pts[1].setX(col+1);		pts[1].setY(row-1);
	pts[2].setX(col+1);		pts[2].setY(row);
	pts[3].setX(col+1);		pts[3].setY(row+1);
	pts[4].setX(col);		pts[4].setY(row+1);
	pts[5].setX(col-1);		pts[5].setY(row+1);
	pts[6].setX(col-1);		pts[6].setY(row);
	pts[7].setX(col-1);		pts[7].setY(row-1);
	
	totalweight = 0.0;
	nearweight = 1.0;
	farweight = 1.0 / 1.4142;
	
	for (int i = 0; i < 8; i++)
	{
		if  ( image.valid( pts[i].x(),pts[i].y() ) && 
					!isHotPixel(pts[i].x(), pts[i].y()) )
		{
			interpolatedrgb = image.pixel(pts[i].x(), pts[i].y());
			
			if ( (i % 2) == 0)
			{
				rsum += ((float)(qRed(interpolatedrgb))) * nearweight;
				gsum += ((float)(qGreen(interpolatedrgb))) * nearweight;
				bsum += ((float)(qBlue(interpolatedrgb))) * nearweight;
				totalweight = totalweight + nearweight;
			}
			else
			{
				rsum += ((float)(qRed(interpolatedrgb))) * farweight;
				gsum += ((float)(qGreen(interpolatedrgb))) * farweight;
				bsum += ((float)(qBlue(interpolatedrgb))) * farweight;
				totalweight = totalweight + farweight;
			}
		}
	}
	
	if (totalweight > 0)
	{
		//round up, otherwise interpolated pixel will appear very slightly,
		//but noticeably darker than the surrounding area.
		rsum = (rsum / totalweight) + 0.5;
		gsum = (gsum / totalweight) + 0.5;
		bsum = (bsum / totalweight) + 0.5;
	}
	else 
	{
		ERROR(ERR_NOINTERPOLATE, "hotpixel::interpolate(), totalweight <= 0");
		return ERR_NOINTERPOLATE;
	}
	
	rgb = qRgb((int)rsum, (int)gsum, (int)bsum);

	return ERR_NOERR;
}

bool hotpixel::isHotPixel(int col, int row)
{
	//for each item in pPixelList call getX() and getY()
	for (int i = 0; i < pixelList.size(); i++)
	{
		if (!pixelList.at(i).isEmpty())
		{
			QString s = pixelList.at(i);
			int x = getX(s);
			int y = getY(s);
			if (x == col && y == row && doneList.at(i) == "false") 
			{	
				return true;
			}	
		}
	}
	
	return false;
}

//These two functions takes a QString object containing an x,y coordinate
//pair and returns the x or y integer component
int hotpixel::getX(QString &s)
{
	QString sx = s.section(',',0,0);
	
	if (!s.isEmpty())
	{
		return atoi(sx.toLatin1().data());
	}
	
	return -1;	
}

int hotpixel::getY(QString &s)
{
	QString sy = s.section(',',1,1);
	
	if (!s.isEmpty())
	{
		return atoi(sy.toLatin1().data());
	}
	
	return -1;	
}

//We keep track of which hotpixels have been interpolated so that
//only the ones which have are included in the interpolation. This 
//gives more accurate results when there are adjacent or nearly
//adjacent hotpixels.
//I know it's stupid to make this a string list, but I need to
//build this on qt3 and qt4 and neither QValueList or QList
//are available on both. Thats why. :)
int hotpixel::initDoneList()
{
	doneList.clear();
	for (int i = 0; i < pixelList.size(); i++)
	{
		doneList.append("false");
	}
	
	return ERR_NOERR;
}

		
//static function
//////////////////////////////////////////////////////////////////////

int hotpixel::getHotPixelList( QString & sListPath, QStringList & slist )
{
	slist.clear();
	
	//open default.hpx in the current working directory
	QFile file(sListPath);
	if ( !file.open(QIODevice::ReadOnly | QIODevice::Text) )
	{
		return ERR_NOFILE;
	}
	
	//read the contents into slist; one line per QString object
	QTextStream in(&file);
	while (!in.atEnd())
	{
		QString line = in.readLine();
		slist.append(line);
	}
	
	return ERR_NOERR;	
}




